# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from typing import (
    Callable,
    Optional,
    Union,
    overload,
)

import numpy as np
from qctrlcommons.node.wrapper import Operation

from boulderopal._nodes.base import create_operation
from boulderopal._nodes.validation import (
    Checker,
    ShapeT,
    number_or_shapeable,
    real_array,
    real_shapeable,
)
from boulderopal._typing import Annotated
from boulderopal._validation import (
    pipe,
    validated,
)

from .node_data import (
    Pwc,
    Stf,
    Tensor,
)
from .utils import (
    get_arraylike_shape,
    get_broadcasted_shape,
    mesh_pwc_durations,
    validate_function_output_shapes,
    validate_tensor_and_function_output_shapes,
)


def _default_broadcast_validator(
    shape_x: tuple[int, ...],
    shape_y: tuple[int, ...],
    x: str,
    y: str,
) -> tuple[int, ...]:
    return get_broadcasted_shape(
        shape_x,
        shape_y,
        message=f"The shapes of {x} and {y} are not broadcastable.",
    )


def _create_flexible_binary_node_data(
    _operation: Operation,
    op_name: str,
    x: int | float | complex | np.ndarray | Tensor | Pwc | Stf,
    y: int | float | complex | np.ndarray | Tensor | Pwc | Stf,
    name: Optional[str],
    validate_value_shape: Callable = _default_broadcast_validator,
) -> Tensor | Pwc | Stf:
    """
    Create the corresponding dataclass for nodes acting on Tensors, Pwcs, and Stfs
    implementing binary functions.

    Parameters
    ----------
    _operation : Operation
        The operation to implement.
    op_name : str
        The name of the operation.
    x : number or np.ndarray or Tensor or Pwc or Stf
        The left operand.
    y : number or np.ndarray or Tensor or Pwc or Stf
        The right operand.
    name : str or None
        The name of the node.
    validate_value_shape : Callable[[tuple, tuple, str, str], tuple], optional
        Function that takes the value shapes of two Tensors, Pwcs,
        or Stfs (as well as their names), and returns the expected values
        shape of the output Tensor, Pwc, or Stf. The function
        shouldn't assume that the shapes are compatible, and raise an
        exception if they aren't. The names provided should be used to
        generate the error message.

    Returns
    -------
    Tensor or Pwc or Stf
        The operation acting on the object.
    """
    # operation(Pwc, Stf) or operation(Stf, Pwc)
    Checker.TYPE(
        not (isinstance(x, Pwc) and isinstance(y, Stf))
        and not (isinstance(x, Stf) and isinstance(y, Pwc)),
        f"You can't apply the {op_name} operation between a Pwc and an Stf.",
        {"x": x, "y": y},
    )

    # operation(Stf, Stf)
    if isinstance(x, Stf) and isinstance(y, Stf):
        Checker.VALUE(name is None, "You can't assign a name to an Stf node.", {"name": name})

        batch_shape, value_shape = validate_function_output_shapes(
            x.batch_shape,
            x.value_shape,
            y.batch_shape,
            y.value_shape,
            validate_value_shape=validate_value_shape,
        )
        return Stf(_operation, value_shape=value_shape, batch_shape=batch_shape)

    # operation(Pwc, Pwc)
    if isinstance(x, Pwc) and isinstance(y, Pwc):
        Checker.VALUE(
            np.isclose(np.sum(x.durations), np.sum(y.durations)),
            "Both Pwc terms must have the same total duration.",
            {"x": x, "y": y},
        )

        batch_shape, value_shape = validate_function_output_shapes(
            x.batch_shape,
            x.value_shape,
            y.batch_shape,
            y.value_shape,
            validate_value_shape=validate_value_shape,
        )

        durations = mesh_pwc_durations([x, y])
        return Pwc(
            _operation,
            value_shape=value_shape,
            durations=durations,
            batch_shape=batch_shape,
        )

    # operation(Stf, ArrayLike) or operation(ArrayLike, Stf)
    if isinstance(x, Stf) or isinstance(y, Stf):
        Checker.VALUE(name is None, "You can't assign a name to an Stf node.", {"name": name})

        if isinstance(x, Stf):
            f_batch_shape, f_value_shape = x.batch_shape, x.value_shape
            f_name = "x"
            t_shape = get_arraylike_shape(y)
            t_name = "y"
            tensor_first = False
        else:
            assert isinstance(y, Stf)
            t_shape = get_arraylike_shape(x)
            t_name = "x"
            f_batch_shape, f_value_shape = y.batch_shape, y.value_shape
            f_name = "y"
            tensor_first = True

        batch_shape, value_shape = validate_tensor_and_function_output_shapes(
            t_shape,
            f_batch_shape,
            f_value_shape,
            t_name,
            f_name,
            validate_value_shape=validate_value_shape,
            tensor_first=tensor_first,
        )

        return Stf(_operation, value_shape=value_shape, batch_shape=batch_shape)

    # operation(Pwc, ArrayLike) or operation(ArrayLike, Pwc)
    if isinstance(x, Pwc) or isinstance(y, Pwc):
        if isinstance(x, Pwc):
            f_batch_shape, f_value_shape = x.batch_shape, x.value_shape
            f_name = "x"
            durations = x.durations
            t_shape = get_arraylike_shape(y)
            t_name = "y"
            tensor_first = False
        else:
            assert isinstance(y, Pwc)
            t_shape = get_arraylike_shape(x)
            t_name = "x"
            f_batch_shape, f_value_shape = y.batch_shape, y.value_shape
            f_name = "y"
            durations = y.durations
            tensor_first = True

        batch_shape, value_shape = validate_tensor_and_function_output_shapes(
            t_shape,
            f_batch_shape,
            f_value_shape,
            t_name,
            f_name,
            validate_value_shape=validate_value_shape,
            tensor_first=tensor_first,
        )

        return Pwc(
            _operation,
            value_shape=value_shape,
            durations=durations,
            batch_shape=batch_shape,
        )

    # operation(ArrayLike, ArrayLike)
    shape = validate_value_shape(get_arraylike_shape(x), get_arraylike_shape(y), "x", "y")
    return Tensor(_operation, shape=shape)


class ArithmeticBinaryGraph:
    """
    Base class implementing binary arithmetic graph methods.
    """

    @overload
    def add(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        y: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def add(
        self,
        x: Pwc,
        y: Union[float, complex, np.ndarray, Tensor, Pwc],
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def add(
        self,
        x: Union[float, complex, np.ndarray, Tensor, Pwc],
        y: Pwc,
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def add(
        self,
        x: Stf,
        y: Union[float, complex, np.ndarray, Tensor, Stf],
    ) -> Stf: ...
    @overload
    def add(
        self,
        x: Union[float, complex, np.ndarray, Tensor, Stf],
        y: Stf,
    ) -> Stf: ...

    @validated
    def add(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        y: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        """
        Calculate the element-wise sum between numbers, np.ndarrays, Tensors, Pwcs,
        or Stfs. You can also use the arithmetic operator ``+`` to calculate their sum.

        Considering numbers and np.ndarrays as Tensors, if the two objects are of the same type,
        so is the returned object. If the objects have different types, Pwcs and Stfs can operate
        with a tensor (returning a Pwc or Stf, respectively).

        This operation supports broadcasting between the different objects.
        When operating a tensor-like object with an Stf or a Pwc, the time dimension of the
        latter is ignored.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The left summand, :math:`x`.
        y : number or np.ndarray or Tensor or Pwc or Stf
            The right summand, :math:`y`.
        name : str or None, optional
            The name of the node. You can only provide a name if neither `x` nor `y` are Stfs.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise sum :math:`x+y`.
        """
        operation = create_operation(self.add, locals())
        return _create_flexible_binary_node_data(operation, "add", x, y, name)

    @overload
    def subtract(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        y: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def subtract(
        self,
        x: Pwc,
        y: Union[float, complex, np.ndarray, Tensor, Pwc],
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def subtract(
        self,
        x: Union[float, complex, np.ndarray, Tensor, Pwc],
        y: Pwc,
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def subtract(
        self,
        x: Stf,
        y: Union[float, complex, np.ndarray, Tensor, Stf],
    ) -> Stf: ...
    @overload
    def subtract(
        self,
        x: Union[float, complex, np.ndarray, Tensor, Stf],
        y: Stf,
    ) -> Stf: ...

    @validated
    def subtract(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        y: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        """
        Calculate the element-wise difference between numbers, np.ndarrays, Tensors, Pwcs,
        or Stfs. You can also use the arithmetic operator ``-`` to calculate their difference.

        Considering numbers and np.ndarrays as Tensors, if the two objects are of the same type,
        so is the returned object. If the objects have different types, Pwcs and Stfs can operate
        with a tensor (returning a Pwc or Stf, respectively).

        This operation supports broadcasting between the different objects.
        When operating a tensor-like object with an Stf or a Pwc, the time dimension of the
        latter is ignored.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The minuend, :math:`x`.
        y : number or np.ndarray or Tensor or Pwc or Stf
            The subtrahend, :math:`y`.
        name : str or None, optional
            The name of the node. You can only provide a name if neither `x` nor `y` are Stfs.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise difference :math:`x-y`.
        """
        operation = create_operation(self.subtract, locals())
        return _create_flexible_binary_node_data(operation, "subtract", x, y, name)

    @overload
    def multiply(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        y: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def multiply(
        self,
        x: Pwc,
        y: Union[float, complex, np.ndarray, Tensor, Pwc],
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def multiply(
        self,
        x: Union[float, complex, np.ndarray, Tensor, Pwc],
        y: Pwc,
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def multiply(
        self,
        x: Stf,
        y: Union[float, complex, np.ndarray, Tensor, Stf],
    ) -> Stf: ...
    @overload
    def multiply(
        self,
        x: Union[float, complex, np.ndarray, Tensor, Stf],
        y: Stf,
    ) -> Stf: ...

    @validated
    def multiply(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        y: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise product between numbers, np.ndarrays, Tensors,
        Pwcs, or Stfs. You can also use the arithmetic operator ``*`` to calculate their product.

        Considering numbers and np.ndarrays as Tensors, if the two objects are of the same type,
        so is the returned object. If the objects have different types, Pwcs and Stfs can operate
        with a tensor (returning a Pwc or Stf, respectively).

        This operation supports broadcasting between the different objects.
        When operating a tensor-like object with an Stf or a Pwc, the time dimension of the
        latter is ignored.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The left factor, :math:`x`.
        y : number or np.ndarray or Tensor or Pwc or Stf
            The right factor, :math:`y`.
        name : str or None, optional
            The name of the node. You can only provide a name if neither `x` nor `y` are Stfs.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise product :math:`x \times y`.
        """
        operation = create_operation(self.multiply, locals())
        return _create_flexible_binary_node_data(operation, "multiply", x, y, name)

    @overload
    def truediv(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        y: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def truediv(
        self,
        x: Pwc,
        y: Union[float, complex, np.ndarray, Tensor, Pwc],
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def truediv(
        self,
        x: Union[float, complex, np.ndarray, Tensor, Pwc],
        y: Pwc,
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def truediv(
        self,
        x: Stf,
        y: Union[float, complex, np.ndarray, Tensor, Stf],
    ) -> Stf: ...
    @overload
    def truediv(
        self,
        x: Union[float, complex, np.ndarray, Tensor, Stf],
        y: Stf,
    ) -> Stf: ...

    @validated
    def truediv(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        y: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        """
        Calculate the element-wise division between numbers, np.ndarrays, Tensors, Pwcs,
        or Stfs. You can also use the arithmetic operator ``/`` to calculate their division.

        Considering numbers and np.ndarrays as Tensors, if the two objects are of the same type,
        so is the returned object. If the objects have different types, Pwcs and Stfs can operate
        with a tensor (returning a Pwc or Stf, respectively).

        This operation supports broadcasting between the different objects.
        When operating a tensor-like object with an Stf or a Pwc, the time dimension of the
        latter is ignored.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The numerator, :math:`x`.
        y : number or np.ndarray or Tensor or Pwc or Stf
            The denominator, :math:`y`.
        name : str or None, optional
            The name of the node. You can only provide a name if neither `x` nor `y` are Stfs.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise division :math:`x/y`.

        See Also
        --------
        Graph.floordiv : Divide two values and take the floor of the result.
        """
        operation = create_operation(self.truediv, locals())
        return _create_flexible_binary_node_data(operation, "truediv", x, y, name)

    @overload
    def floordiv(
        self,
        x: Union[float, np.ndarray, Tensor],
        y: Union[float, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def floordiv(
        self,
        x: Pwc,
        y: Union[float, np.ndarray, Tensor, Pwc],
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def floordiv(
        self,
        x: Union[float, np.ndarray, Tensor, Pwc],
        y: Pwc,
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def floordiv(
        self,
        x: Stf,
        y: Union[float, np.ndarray, Tensor, Stf],
    ) -> Stf: ...
    @overload
    def floordiv(
        self,
        x: Union[float, np.ndarray, Tensor, Stf],
        y: Stf,
    ) -> Stf: ...

    @validated
    def floordiv(
        self,
        x: Annotated[
            Union[float, np.ndarray, Tensor, Pwc, Stf],
            pipe(real_shapeable, after=real_array),
        ],
        y: Annotated[
            Union[float, np.ndarray, Tensor, Pwc, Stf],
            pipe(real_shapeable, after=real_array),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        r"""
        Calculate the element-wise rounded-down division between real numbers, np.ndarrays, Tensors,
        Pwcs, or Stfs. You can also use the arithmetic operator ``//`` to calculate their floor
        division.

        Considering numbers and np.ndarrays as Tensors, if the two objects are of the same type,
        so is the returned object. If the objects have different types, Pwcs and Stfs can operate
        with a tensor (returning a Pwc or Stf, respectively).

        This operation supports broadcasting between the different objects.
        When operating a tensor-like object with an Stf or a Pwc, the time dimension of the
        latter is ignored.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The numerator, :math:`x`. Must be real-valued.
        y : number or np.ndarray or Tensor or Pwc or Stf
            The denominator, :math:`y`. Must be real-valued.
        name : str or None, optional
            The name of the node. You can only provide a name if neither `x` nor `y` are Stfs.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise rounded-down division :math:`\lfloor x/y \rfloor`.

        See Also
        --------
        Graph.truediv : Divide two values.
        """
        operation = create_operation(self.floordiv, locals())
        return _create_flexible_binary_node_data(operation, "floordiv", x, y, name)

    @overload
    def pow(
        self,
        x: Union[float, complex, np.ndarray, Tensor],
        y: Union[float, complex, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def pow(
        self,
        x: Pwc,
        y: Union[float, complex, np.ndarray, Tensor, Pwc],
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def pow(
        self,
        x: Union[float, complex, np.ndarray, Tensor, Pwc],
        y: Pwc,
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def pow(
        self,
        x: Stf,
        y: Union[float, complex, np.ndarray, Tensor, Stf],
    ) -> Stf: ...
    @overload
    def pow(
        self,
        x: Union[float, complex, np.ndarray, Tensor, Stf],
        y: Stf,
    ) -> Stf: ...

    @validated
    def pow(
        self,
        x: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        y: Annotated[
            Union[float, complex, np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        """
        Calculate the element-wise power between numbers, np.ndarrays, Tensors, Pwcs,
        or Stfs. You can also use the arithmetic operator ``**`` to calculate their power.

        Considering numbers and np.ndarrays as Tensors, if the two objects are of the same type,
        so is the returned object. If the objects have different types, Pwcs and Stfs can operate
        with a tensor (returning a Pwc or Stf, respectively).

        This operation supports broadcasting between the different objects.
        When operating a tensor-like object with an Stf or a Pwc, the time dimension of the
        latter is ignored.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The base, :math:`x`.
        y : number or np.ndarray or Tensor or Pwc or Stf
            The exponent, :math:`y`.
        name : str or None, optional
            The name of the node. You can only provide a name if neither `x` nor `y` are Stfs.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise power :math:`x^y`.

        Warnings
        --------
        This function considers that the zeroth power of zero (:math:`0^0`) is
        undefined. This means that you might see an error if you attempt to
        fetch an object that contains :math:`0^0`.
        """
        operation = create_operation(self.pow, locals())
        return _create_flexible_binary_node_data(operation, "pow", x, y, name)

    @overload
    def complex_value(
        self,
        x: Union[float, np.ndarray, Tensor],
        y: Union[float, np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def complex_value(
        self,
        x: Pwc,
        y: Union[float, np.ndarray, Tensor, Pwc],
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def complex_value(
        self,
        x: Union[float, np.ndarray, Tensor, Pwc],
        y: Pwc,
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def complex_value(
        self,
        x: Stf,
        y: Union[float, np.ndarray, Tensor, Stf],
    ) -> Stf: ...
    @overload
    def complex_value(
        self,
        x: Union[float, np.ndarray, Tensor, Stf],
        y: Stf,
    ) -> Stf: ...

    @validated
    def complex_value(
        self,
        x: Annotated[
            Union[float, np.ndarray, Tensor, Pwc, Stf],
            pipe(real_shapeable, after=real_array),
        ],
        y: Annotated[
            Union[float, np.ndarray, Tensor, Pwc, Stf],
            pipe(real_shapeable, after=real_array),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        """
        Create element-wise complex values from real numbers, np.ndarrays, Tensors,
        Pwcs, or Stfs, that is, the real and imaginary parts.

        Considering numbers and np.ndarrays as Tensors, if the two objects are of the same type,
        so is the returned object. If the objects have different types, Pwcs and Stfs can operate
        with a tensor (returning a Pwc or Stf, respectively).

        This operation supports broadcasting between the different objects.
        When operating a tensor-like object with an Stf or a Pwc, the time dimension of the
        latter is ignored.

        Parameters
        ----------
        x : number or np.ndarray or Tensor or Pwc or Stf
            The real part, :math:`x`.
        y : number or np.ndarray or Tensor or Pwc or Stf
            The imaginary part, :math:`y`.
        name : str or None, optional
            The name of the node. You can only provide a name if neither `x` nor `y` are Stfs.

        Returns
        -------
        Tensor or Pwc or Stf
            The element-wise complex number :math:`x+iy`.

        See Also
        --------
        Graph.abs : Absolute value of a complex object.
        Graph.angle : Argument of a complex object.
        Graph.conjugate : Conjugate of a complex object.
        Graph.imag : Imaginary part of a complex object.
        Graph.real : Real part of a complex object.
        """
        operation = create_operation(self.complex_value, locals())
        return _create_flexible_binary_node_data(operation, "complex_value", x, y, name)

    @overload
    def matmul(
        self,
        x: Union[np.ndarray, Tensor],
        y: Union[np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def matmul(
        self,
        x: Pwc,
        y: Union[np.ndarray, Tensor, Pwc],
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def matmul(
        self,
        x: Union[np.ndarray, Tensor, Pwc],
        y: Pwc,
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def matmul(
        self,
        x: Stf,
        y: Union[np.ndarray, Tensor, Stf],
    ) -> Stf: ...
    @overload
    def matmul(
        self,
        x: Union[np.ndarray, Tensor, Stf],
        y: Stf,
    ) -> Stf: ...

    @validated
    def matmul(
        self,
        x: Annotated[
            Union[np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable, after=ShapeT.MATRIX()),
        ],
        y: Annotated[
            Union[np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable, after=ShapeT.MATRIX()),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        """
        Calculate the matrix multiplication between np.ndarrays, Tensors,
        Pwcs, or Stfs. You can also use the arithmetic operator ``@``
        to calculate their matrix multiplication.

        If any of the inputs is a Pwc or Stf, the output is also a
        Pwc or Stf (mixing Pwcs and Stfs is not supported).
        Otherwise, the output is a Tensor.

        This operation supports broadcasting between the batch dimensions of
        the two input objects. All the dimensions with the exception of the two
        innermost ones (where the matrix multiplication is performed) are
        considered batch dimensions.

        When operating a tensor-like object with an Stf or a Pwc, the time
        dimension of the latter is ignored.

        Parameters
        ----------
        x : np.ndarray or Tensor or Pwc or Stf
            The left multiplicand. It must be a matrix (or batch of matrices)
            and its last dimension must be the same as the second-to-last
            dimension of `y`.
        y : np.ndarray or Tensor or Pwc or Stf
            The right multiplicand. It must be a matrix (or batch of matrices)
            and its second-to-last dimension must be the same as the last
            dimension of `x`.
        name : str or None, optional
            The name of the node. You can only provide a name if neither `x`
            nor `y` are Stfs.

        Returns
        -------
        Tensor or Pwc or Stf
            The matrix product of the input objects. If any of the input
            objects is a Pwc or Stf, the returned objects has the same
            type. Otherwise, it is a Tensor.

        See Also
        --------
        Graph.einsum : Tensor contraction via Einstein summation convention.
        """

        def validate_value_shape(
            x_shape: tuple[int, ...],
            y_shape: tuple[int, ...],
            x_name: str,
            y_name: str,
        ) -> tuple[int, ...]:
            trailing_shape = _default_broadcast_validator(
                x_shape[:-2],
                y_shape[:-2],
                x_name,
                y_name,
            )
            Checker.VALUE(
                x_shape[-1] == y_shape[-2],
                "The last dimension of x must be equal to the second-to-last dimension of y.",
                {"x shape": x_shape, "y shape": y_shape},
            )

            return trailing_shape + (x_shape[-2], y_shape[-1])

        operation = create_operation(self.matmul, locals())
        return _create_flexible_binary_node_data(
            operation,
            "matmul",
            x,
            y,
            name,
            validate_value_shape,
        )

    @overload
    def kron(
        self,
        x: Union[np.ndarray, Tensor],
        y: Union[np.ndarray, Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor: ...
    @overload
    def kron(
        self,
        x: Pwc,
        y: Union[np.ndarray, Tensor, Pwc],
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def kron(
        self,
        x: Union[np.ndarray, Tensor, Pwc],
        y: Pwc,
        *,
        name: Optional[str] = None,
    ) -> Pwc: ...
    @overload
    def kron(
        self,
        x: Stf,
        y: Union[np.ndarray, Tensor, Stf],
    ) -> Stf: ...
    @overload
    def kron(
        self,
        x: Union[np.ndarray, Tensor, Stf],
        y: Stf,
    ) -> Stf: ...

    @validated
    def kron(
        self,
        x: Annotated[
            Union[np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable, after=ShapeT.MATRIX()),
        ],
        y: Annotated[
            Union[np.ndarray, Tensor, Pwc, Stf],
            pipe(number_or_shapeable, after=ShapeT.MATRIX()),
        ],
        *,
        name: Optional[str] = None,
    ) -> Union[Tensor, Pwc, Stf]:
        """
        Calculate the Kronecker product between np.ndarrays, Tensors,
        Pwcs, or Stfs.

        If any of the inputs is a Pwc or Stf, the output is also a
        Pwc or Stf (mixing Pwcs and Stfs is not supported).
        Otherwise, the output is a Tensor.

        This operation supports broadcasting between the batch dimensions of
        the two input objects. All the dimensions with the exception of the two
        innermost ones (where the Kronecker product is performed) are
        considered batch dimensions.

        When operating a tensor-like object with an Stf or a Pwc, the time
        dimension of the latter is ignored.

        Parameters
        ----------
        x : np.ndarray or Tensor or Pwc or Stf
            The left multiplicand. It must be a have at least two dimensions.
        y : np.ndarray or Tensor or Pwc or Stf
            The right multiplicand. It must be a have at least two dimensions.
        name : str or None, optional
            The name of the node. You can only provide a name if neither `x`
            nor `y` are Stfs.

        Returns
        -------
        Tensor or Pwc or Stf
            The Kronecker product of the input objects. If any of the input
            objects is a Pwc or Stf, the returned objects has the same
            type. Otherwise, it is a Tensor.

        See Also
        --------
        Graph.embed_operators : Embed operators into a larger Hilbert space.
        Graph.kronecker_product_list : Kronecker product of a list of operators.
        Graph.pauli_kronecker_product : Embed Pauli matrices into a larger Hilbert space.
        """

        def validate_value_shape(
            x_shape: tuple[int, ...],
            y_shape: tuple[int, ...],
            x_name: str,
            y_name: str,
        ) -> tuple[int, ...]:
            trailing_shape = _default_broadcast_validator(
                x_shape[:-2],
                y_shape[:-2],
                x_name,
                y_name,
            )
            return trailing_shape + (
                x_shape[-2] * y_shape[-2],
                x_shape[-1] * y_shape[-1],
            )

        operation = create_operation(self.kron, locals())
        return _create_flexible_binary_node_data(
            operation,
            "kron",
            x,
            y,
            name,
            validate_value_shape,
        )



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from qctrlworkflowclient.router.api import (
    Action,
    ActionStatus,
)

from boulderopal._core.formatter import metadata_formatter
from boulderopal._validation import Checker
from boulderopal.cloud._utils import get_cloud_router


class BoulderOpalJob:
    """
    An object tracking a Boulder Opal calculation submitted to the cloud.

    Parameters
    ----------
    action : Action
        The unique identifier object for the cloud job.

    Attributes
    ----------
    action_id : str
        The ID associated with the calculation.
    """

    def __init__(self, action: Action) -> None:
        self._action = action
        self._router = get_cloud_router("Manipulating a BoulderOpalJob")

    def __repr__(self) -> str:
        return f'<BoulderOpalJob: action_id="{self.action_id}">'

    @property
    def action_id(self) -> str:
        """
        The job action ID.
        """
        return self._action.action_id

    def get_status(self) -> str:
        """
        Get the job status.

        Returns
        -------
        str
            The status of the job.
        """
        self._action = self._router.update_action_status(action=self._action)
        assert self._action.status is not None, "Unknown action status."
        return self._action.status

    def get_result(self) -> dict:
        """
        Get the job result.
        Note that this is a blocking operation.

        Returns
        -------
        dict
            A dictionary containing the calculation result.
        """
        return metadata_formatter(
            self._router.get_result(self._action, revoke_on_interrupt=False),
        ).decoded

    def cancel(self) -> None:
        """
        Cancel the remote job.
        """
        status = self.get_status()
        if status == ActionStatus.REVOKED.value:
            message = "was already cancelled"
        elif status == ActionStatus.SUCCESS.value:
            message = "has already completed"
        elif status == ActionStatus.FAILURE.value:
            message = "failed"
        else:
            message = ""

        if message != "":
            print(
                f'Your task (action_id="{self.action_id}") {message}, and cannot be cancelled.',
            )
        else:
            self._router._revoke_action(
                self._action,
            )
            print(f'Your task (action_id="{self.action_id}") has been cancelled.')


def get_job(action_id: int | str) -> BoulderOpalJob:
    """
    Create an object to track a remote job from its action ID.

    Parameters
    ----------
    action_id : int or str
        The ID associated with the calculation.
        You can get the ID of your calculations from the activity monitor.

    Returns
    -------
    BoulderOpalJob
        The object tracking the remote job.
    """
    Checker.TYPE(
        str(action_id).isnumeric(),
        "The calculation ID must be an integer.",
        {"action_id": action_id},
    )
    return BoulderOpalJob(Action(action_id=str(action_id)))



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import (
    List,
    Optional,
)

import numpy as np

from boulderopal._nodes.base import create_operation
from boulderopal._nodes.node_data import Tensor
from boulderopal._validation import (
    Checker,
    validated,
)


class DifferentiationGraph:
    """
    Base class implementing differentiation graph methods.
    """

    @validated
    def hessian(
        self,
        tensor: Tensor,
        variables: List[Tensor],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate a single Hessian matrix for all the variables.

        The Hessian is a matrix containing all the second partial derivatives
        of the `tensor` with respect to the `variables`.

        Parameters
        ----------
        tensor : Tensor
            The real scalar tensor :math:`T` whose Hessian matrix you want to
            calculate.
        variables : list[Tensor]
            The list of real variables :math:`\{\theta_i\}` with respect to
            which you want to take the second partial derivatives of the
            tensor. If any of the tensors of the list is not scalar, this
            function treats each of the elements of the tensor as a different
            variable. It does this by flattening all tensors and concatenating
            them in the same sequence that you provided in this list.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            A 2D real Hessian matrix :math:`H` containing the second partial
            derivatives of the `tensor` :math:`T` with respect to the
            `variables` :math:`\{\theta_i\}`.

        Warnings
        --------
        This function currently doesn't support calculating a Hessian matrix for
        a graph which includes an `infidelity_pwc` node if it involves a Hamiltonian
        with degenerate eigenvalues at any segment. In that case, the function
        returns a `nan` Hessian matrix.

        Notes
        -----
        The :math:`(i,j)` element of the Hessian contains the partial
        derivative of the `tensor` with respect to the ith and the jth
        `variables`:

        .. math::
            H_{i,j} = \frac{\partial^2 T}{\partial \theta_i \partial \theta_j}.

        The variables :math:`\{\theta_i\}` follow the same sequence as the
        input list of `variables`. If some of the `variables` are not scalars,
        this function flattens them and concatenates them in the same order of
        the list of `variables` that you provided to create the sequence of
        scalar variables :math:`\{\theta_i\}`.

        If you choose a negative log-likelihood as the tensor :math:`T`, you
        can use this Hessian as an estimate of the Fisher information matrix.
        """
        Checker.VALUE(tensor.operation.is_scalar_tensor, "The tensor must be a scalar tensor.")
        variable_count = int(np.sum([np.prod(variable.shape) for variable in variables]))
        shape = (variable_count, variable_count)
        operation = create_operation(self.hessian, locals())
        return Tensor(operation, shape=shape)



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import (
    TYPE_CHECKING,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np

from boulderopal._nodes.base import create_operation
from boulderopal._nodes.node_data import Sequence as Sequence_
from boulderopal._nodes.node_data import (
    Stf,
    Tensor,
)
from boulderopal._nodes.validation import shapeable
from boulderopal._typing import Annotated
from boulderopal._validation import (
    Checker,
    ScalarT,
    pipe,
    validated,
)

if TYPE_CHECKING:
    from boulderopal.graph import Graph


class RandomGraph:
    """
    Base class implementing random graph methods.
    """

    def __init__(self, graph: "Graph") -> None:
        self._graph = graph

    @validated
    def colored_noise_stf_signal(
        self,
        power_spectral_density: Annotated[Union[np.ndarray, Tensor], pipe(shapeable)],
        frequency_step: Annotated[float, pipe(ScalarT.REAL().gt(0))],
        batch_shape: Tuple[Annotated[int, pipe(ScalarT.INT().gt(0))], ...] = (),
        seed: Optional[Annotated[int, pipe(ScalarT.INT())]] = None,
    ) -> Stf:
        r"""
        Sample the one-sided power spectral density (PSD) of a random noise process in the
        time domain and returns the resultant noise trajectories as an Stf.

        Parameters
        ----------
        power_spectral_density : np.ndarray or Tensor (1D, real)
            The one-sided power spectral density of the noise sampled at frequencies
            :math:`\{0, \Delta f, 2\Delta f, \ldots , M\Delta f\}`.
        frequency_step : float
            The step size :math:`\Delta f` of power spectrum densities samples
           `power_spectral_density`. Must be a strictly positive number.
        batch_shape : list[int] or tuple[int], optional
            The batch shape of the returned Stf. By default, the batch shape is ``()``, that is,
            the returned Stf represents only one noise trajectory. If the batch shape is
            ``(m, n,...)``, the returned Stf represents `m*n*...` trajectories arranged in
            this batch shape.
        seed : int or None, optional
            A seed for the random number generator used for sampling. When set, same
            trajectories are produced on every run of this function, provided all the other
            arguments also remain unchanged. Defaults to None, in which case the
            generated noise trajectories can be different from one run to another.

        Returns
        -------
        Stf
            An `Stf` signal representing the noise trajectories in the time domain. The
            batch shape of this `Stf` is same as the argument `batch_shape`.

        Notes
        -----
        Given a frequency step size of :math:`\Delta f` and discrete samples
        :math:`P[k] = P(k\Delta f)` of a one-sided power spectral density function
        :math:`P(f)`, the output is a possibly batched Stf which represents one random
        realization of the random noise process. Each such trajectory is periodic with a
        time period of :math:`1/\Delta f`.

        Examples
        --------
        Create a PWC signal by sampling from a power spectral density function
        to define a noise Hamiltonian.

        >>> sigma_z = np.diag([1, -1])
        >>> frequency_step = 2e3
        >>> frequencies = np.arange(0, 2e6, frequency_step)
        >>> frequency_cutoff = 0.05e6
        >>> power_densities = 4e9 / (frequencies + frequency_cutoff)
        >>> noise_stf = graph.random.colored_noise_stf_signal(
        ...     power_spectral_density=power_densities, frequency_step=2000.0, batch_shape=(100,)
        ... )
        >>> noise_stf
        <Stf: operation_name="random_colored_noise_stf_signal", value_shape=(), batch_shape=(100,)>
        >>> noise_pwc = graph.discretize_stf(stf=noise_stf, duration=2e-6, segment_count=50)
        >>> noise_hamiltonian = noise_pwc * sigma_z
        >>> noise_hamiltonian
        <Pwc: name="multiply_#3", operation_name="multiply", value_shape=(2, 2), batch_shape=(100,)>

        Refer to the `How to simulate quantum dynamics subject to noise with graphs
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-quantum-dynamics-
        subject-to-noise-with-graphs>`_ user guide to find the example in context.
        """
        Checker.VALUE(
            len(power_spectral_density.shape) == 1,
            "The power spectral density must be 1D.",
        )

        operation = create_operation(
            self.colored_noise_stf_signal,
            locals(),
            graph=self._graph,
            name="random_colored_noise_stf_signal",
        )
        return Stf(operation, value_shape=(), batch_shape=batch_shape)

    @validated
    def normal(
        self,
        shape: Tuple[Annotated[int, pipe(ScalarT.INT().gt(0))], ...],
        mean: Annotated[float, pipe(ScalarT.REAL())],
        standard_deviation: Annotated[float, pipe(ScalarT.REAL().gt(0))],
        seed: Optional[Annotated[int, pipe(ScalarT.INT())]] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create a sample of normally distributed random numbers.

        Parameters
        ----------
        shape : tuple[int, ...]
            The shape of the sampled random numbers.
        mean : float
            The mean of the normal distribution.
        standard_deviation : float
            The standard deviation of the normal distribution.
        seed : int or None, optional
            A seed for the random number generator. Defaults to None,
            in which case a random value for the seed is used.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            A tensor containing a sample of normally distributed random numbers
            with shape ``shape``.

        See Also
        --------
        :func:`Graph.random.choices <random.choices>`
            Create random samples from the data that you provide.
        :func:`Graph.random.uniform <random.uniform>`
            Create a sample of uniformly distributed random numbers.
        :func:`boulderopal.run_stochastic_optimization`
            Function to find the minimum of generic stochastic functions.

        Examples
        --------
        Create a random tensor by sampling from a Gaussian distribution.

        >>> samples = graph.random.normal(
        ...     shape=(3, 1), mean=0.0, standard_deviation=0.05, seed=0, name="samples"
        ... )
        >>> result = bo.execute_graph(graph=graph, output_node_names="samples")
        >>> result["output"]["samples"]["value"]
        array([[-0.03171833], [0.00816805], [-0.06874011]])

        Create a batch of noise signals to construct a PWC Hamiltonian. The signal is defined
        as :math:`a \cos(\omega t)`, where :math:`a` follows a normal distribution and
        :math:`\omega` follows a uniform distribution.

        >>> seed = 0
        >>> batch_size = 3
        >>> sigma_x = np.array([[0, 1], [1, 0]])
        >>> sample_times = np.array([0.1, 0.2])
        >>> a = graph.random.normal((batch_size, 1), mean=0.0, standard_deviation=0.05, seed=seed)
        >>> omega = graph.random.uniform(
        ...     shape=(batch_size, 1), lower_bound=np.pi, upper_bound=2 * np.pi, seed=seed
        ... )
        >>> sampled_signal = a * graph.cos(omega * sample_times[None])
        >>> hamiltonian = graph.pwc_signal(sampled_signal, duration=0.2) * sigma_x
        >>> hamiltonian.name = "hamiltonian"
        >>> result = bo.execute_graph(graph=graph, output_node_names="hamiltonian")
        >>> result["output"]["hamiltonian"]
        {
            'durations': array([0.1, 0.1]),
            'values': array([
                [
                    [[-0.        , -0.02674376], [-0.02674376, -0.        ]],
                    [[-0.        , -0.01338043], [-0.01338043, -0.        ]]
                ],
                [
                    [[ 0.        ,  0.00691007], [ 0.00691007,  0.        ]],
                    [[ 0.        ,  0.00352363], [ 0.00352363,  0.        ]]],
                [
                    [[-0.        , -0.06230612], [-0.06230612, -0.        ]],
                    [[-0.        , -0.04420857], [-0.04420857, -0.        ]]
                ]
            ]),
            'time_dimension': 1
        }

        See more examples in the `How to optimize controls robust to strong noise sources
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-robust-
        to-strong-noise-sources>`_ user guide.
        """

        operation = create_operation(self.normal, locals(), self._graph, "random_normal")
        return Tensor(operation, shape=tuple(shape))

    @validated
    def uniform(
        self,
        shape: Tuple[Annotated[int, pipe(ScalarT.INT().gt(0))], ...],
        lower_bound: Annotated[float, pipe(ScalarT.REAL())],
        upper_bound: Annotated[float, pipe(ScalarT.REAL())],
        seed: Optional[Annotated[int, pipe(ScalarT.INT())]] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create a sample of uniformly distributed random numbers.

        Parameters
        ----------
        shape : tuple
            The shape of the sampled random numbers.
        lower_bound : float
            The inclusive lower bound of the interval of the uniform distribution.
        upper_bound : float
            The exclusive upper bound of the interval of the uniform distribution.
        seed : int or None, optional
            A seed for the random number generator. Defaults to None,
            in which case a random value for the seed is used.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            A tensor containing a sample of uniformly distributed random numbers
            with shape ``shape``.

        See Also
        --------
        :func:`Graph.random.choices <random.choices>`
            Create random samples from the data that you provide.
        :func:`Graph.random.normal <random.normal>`
            Create a sample of normally distributed random numbers.
        :func:`boulderopal.run_stochastic_optimization`
            Function to find the minimum of generic stochastic functions.

        Examples
        --------
        Create a random tensor by sampling uniformly from :math:`[0,\, 1)`.

        >>> samples = graph.random.uniform(
        ...     shape=(3, 1), lower_bound=0, upper_bound=1, seed=0, name="samples"
        ... )
        >>> result = bo.execute_graph(graph=graph, output_node_names="samples")
        >>> result["output"]["samples"]["value"]
        array([[0.8069013], [0.79011373], [0.38818516]])

        See more examples in the `How to optimize controls robust to strong noise sources
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-
        robust-to-strong-noise-sources>`_ user guide.
        """
        Checker.VALUE(
            lower_bound < upper_bound,
            "The lower bound must be smaller than the upper bound.",
            {"lower_bound": lower_bound, "upper_bound": upper_bound},
        )

        operation = create_operation(self.uniform, locals(), self._graph, "random_uniform")
        return Tensor(operation, shape=shape)

    @validated
    def choices(
        self,
        data: List[Annotated[Union[np.ndarray, Tensor], pipe(shapeable)]],
        sample_count: Annotated[int, pipe(ScalarT.INT().gt(0))],
        seed: Optional[Annotated[int, pipe(ScalarT.INT())]] = None,
        *,
        name: Optional[str] = None,
    ) -> Sequence[Tensor]:
        r"""
        Create random samples from the data that you provide.

        You can provide the data as a list and each element of that list represents one component
        of the full data. For example, considering a single variable linear regression problem
        that is described by the input :math:`x` and output :math:`y`, the data you provide would
        be :math:`[x, y]`. The first dimension of the data component in this list is the size of
        the data and therefore must be same for all components. However, all these components can
        have different value shapes, meaning the other dimensions can vary.

        This node effectively chooses a random batch of `sample_count` indices :math:`\{s_i\}`,
        and extracts the corresponding slices :math:`\{c[s_i]\}` of each data component.
        For example, in the case of linear regression, you can use this node to extract a random
        subset of your full data set.

        If this node is evaluated multiple times (for example during an optimization), it samples
        indices without replacement until all indices have been seen, at which point it starts
        sampling from the full set of indices again. You can therefore use this node to create
        minibatches that iterate over your data set in a series of epochs.

        Parameters
        ----------
        data : list[np.ndarray or Tensor]
            A list of data components. The first dimensions of the elements in this
            list denote the total amount of the data, and therefore must be the same.
        sample_count : int
            Number of samples in the returned batch.
        seed : int or None, optional
            Seed for random number generator. Defaults to None. If set, it ensures the
            random samples are generated in a reproducible sequence.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Sequence[Tensor]
            A sequence representing a batch of random samples from `data`.
            You can access the elements of the sequence using integer indices.
            The number of elements of the sequence is the same as the size of `data`.
            Each element of the sequence has the length (along its first dimension)
            as defined by `sample_count`.

        See Also
        --------
        :func:`Graph.random.normal <random.normal>`
            Create a sample of normally distributed random numbers.
        :func:`Graph.random.uniform <random.uniform>`
            Create a sample of uniformly distributed random numbers.
        :func:`boulderopal.run_stochastic_optimization`
            Function to find the minimum of generic stochastic functions.

        Examples
        --------
        >>> x = np.arange(20).reshape((10, 2))
        >>> y = np.arange(10) * 0.2
        >>> sampled_x, sampled_y = graph.random.choices([x, y], 3, seed=1)
        >>> sampled_x.name = "sampled_x"
        >>> sampled_y.name = "sampled_y"
        >>> result = bo.execute_graph(graph=graph, output_node_names=["sampled_x", "sampled_y"])
        >>> result["output"]["sampled_x"]["value"]
        array([[2, 3], [6, 7], [0, 1]])
        >>> result["output"]["sampled_y"]["value"]
        array([0.2, 0.6, 0. ])

        See more examples in the `How to perform Hamiltonian parameter estimation using a large
        amount of measured data
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-perform-parameter-estimation-
        with-a-large-amount-of-data>`_ user guide.
        """
        data_size = data[0].shape[0]
        Checker.VALUE(
            all(value.shape[0] == data_size for value in data),
            "The first dimension of the elements in data must be the same.",
            {"data": data},
        )
        Checker.VALUE(
            sample_count <= data_size,
            "The sample_count must be not greater than the size of the data you provide.",
            {"sample_count": sample_count, "data size": data_size},
        )

        return_tensor_shapes = [(sample_count,) + value.shape[1:] for value in data]
        operation = create_operation(self.choices, locals(), self._graph, "random_choices")
        return Sequence_(operation).create_sequence(
            node_constructor=lambda operation, index: Tensor(
                operation,
                return_tensor_shapes[index],
            ),
            size=len(data),
        )



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

import time
import warnings
from typing import (
    Any,
    ContextManager,
)

from qctrlcommons.exceptions import QctrlException
from qctrlworkflowclient import (
    ApiRouter,
    LocalRouter,
    get_authenticated_client_for_product,
)
from qctrlworkflowclient.router.api import (
    Action,
    ActionStatus,
)
from rich.console import Console
from rich.table import Table

from boulderopal._configuration.configuration import (
    configure,
    get_configuration,
)
from boulderopal._constants import PACKAGE_INFO
from boulderopal._core.formatter import (
    async_result_formatter,
    metadata_formatter,
)
from boulderopal._validation import (
    Checker,
    ScalarT,
)


def get_cloud_router(on_fail_message: str) -> ApiRouter:
    """
    Return the current cloud router.

    Parameters
    ----------
    on_fail_message : str
        Message to throw when running locally.

    Returns
    -------
    ApiRouter
        The current ApiRouter object.

    Raises
    ------
    QctrlException
        If running in local mode.
    """
    router = get_configuration().get_router()
    if isinstance(router, LocalRouter):
        raise QctrlException(f"{on_fail_message} is not supported in local mode.")
    return router


def get_result(action_id: int | str) -> dict:
    """
    Retrieve the result of a previously run calculation.

    Parameters
    ----------
    action_id : int or str
        The ID associated to the calculation.
        You can get the ID of your calculations from the activity monitor.

    Returns
    -------
    dict
        A dictionary containing the calculation result.

    Warnings
    --------
    This function is deprecated and will be removed in a future release.
    You can instead use `boulderopal.cloud.get_job` to create a `BoulderOpalJob` object associated
    to the calculation and use its `get_result` method.
    """
    warnings.warn(
        "This function is deprecated and will be removed in a future release. "
        "You can instead use `boulderopal.cloud.get_job` to create a `BoulderOpalJob` object "
        "associated to the calculation and use its `get_result` method.",
        category=FutureWarning,
        stacklevel=2,
    )
    Checker.TYPE(
        str(action_id).isnumeric(),
        "The calculation id must be an integer.",
        {"action_id": action_id},
    )
    action = Action(action_id=str(action_id))
    return metadata_formatter(get_cloud_router("Retrieving results").get_result(action)).decoded


def request_machines(machine_count: int) -> None:
    """
    Request a minimum number of machines to be online.

    Parameters
    ----------
    machine_count : int
        The minimum number of machines requested to be online.

    Notes
    -----
    This command is blocking until the specified amount of machines
    have been observed in your environment. It only attempts to ensure
    the requested amount are online, not necessarily starting up that
    amount if some machines are already online.
    """
    machine_count = ScalarT.INT("machine_count").ge(1)(machine_count)
    get_cloud_router("Requesting machines").request_machines(machine_count)


def shut_down_machines(force: bool = False) -> None:
    """
    Shut down all the machines in your organization's cloud environment.

    Parameters
    ----------
    force : bool, optional
        Whether to unconditionally shut down the machines.
        If set to False (the default), the machines will only be shut down
        if there are no running or queued calculations in your organization's queue.
        If set to True, the machines will be shut down regardless of the status of the calculations.
    """
    router = get_cloud_router("Shutting down machines")
    if not force:
        status = (ActionStatus.PENDING.name, ActionStatus.STARTED.name)
        for s in status:
            action_records: list[dict[str, Any]] = router.activity_monitor(
                limit=99,
                offset=0,
                status=s,
                only_user_results=False,
            )
            if len(action_records) > 0:
                print(
                    "Could not shut down the machines "
                    "as there are still calculations running or in the queue. "
                    "You can try again when they are finished, "
                    "or set the force parameter to True to shut down the machines unconditionally.",
                )
                return

    result = router.shut_down_machines()
    errors = result["terminateTenant"]["errors"]
    if errors:
        print("There was an error while shutting down the machines.")
        print(errors)
    else:
        # To prevent machines from not starting when there's a task in the queue,
        # due to the short time gap between shutdown and task submission.
        time.sleep(20)
        print("All machines have been shut down successfully.")


def group_requests() -> ContextManager:
    """
    Create a context manager for executing multiple function calls over available machines.

    Returns
    -------
    ContextManager
        A context manager to collect and run computation requests.

    Warnings
    --------
    This function is deprecated and will be removed in a future release.
    You can instead submit multiple asynchronous calculations if the functions support it.

    Notes
    -----
    All grouped calculations must be independent from each other.

    Within the context manager, the object returned from each request is a placeholder.
    When exiting, the context manager waits until all calculations have finished,
    hence this command blocks execution.
    When all results are received, the placeholders are replaced with them.

    Read the `Computational resources in Boulder Opal
    <https://docs.q-ctrl.com/boulder-opal/topics/computational-resources-in-boulder-opal>`_
    topic for more information about this feature.
    """
    warnings.warn(
        "This function is deprecated and will be removed in a future release. "
        "You can instead submit multiple asynchronous calculations if the functions support it.",
        category=FutureWarning,
        stacklevel=2,
    )
    return get_cloud_router("Grouping requests").enable_parallel(callback=async_result_formatter)


def set_organization(organization_slug: str) -> None:
    """
    Set the organization for Boulder Opal cloud calls.

    Parameters
    ----------
    organization_slug : str
        Unique slug for the organization.
    """
    configure(organization=organization_slug)


def set_api(api_url: str, oidc_url: str) -> None:
    """
    Configure Boulder Opal for API routing.

    Parameters
    ----------
    api_url : str
        URL of the GraphQL schema.
    oidc_url : str
        Base URL of the OIDC provider, e.g. Keycloak.
    """
    client = get_authenticated_client_for_product(
        package_name=PACKAGE_INFO.install_name,
        api_url=api_url,
        auth=oidc_url,
    )
    settings = get_configuration()

    configure(router=ApiRouter(client, settings))


def show_queue_status(only_user_results: bool = False) -> None:
    """
    Show the number of queued and running calculations
    in your Boulder Opal organization's cloud environment.

    Parameters
    ----------
    only_user_results : bool, optional
        If set to False (the default), all calculations in your organization will be displayed.
        If set to True, only information about your calculations will be displayed.

    Notes
    -----
    This function counts up to 99 results per status,
    even if there are more actions with the same status.
    """
    status_map = {
        ActionStatus.PENDING.name: "Queued",
        ActionStatus.STARTED.name: "Running",
    }
    status = {}
    router = get_cloud_router("Queue status")
    for s in status_map:
        action_records: list[dict[str, Any]] = router.activity_monitor(
            limit=99,
            offset=0,
            status=s,
            only_user_results=only_user_results,
        )
        status[s] = len(action_records)

    table = Table(title="Queue status", show_lines=True)
    console = Console()

    for header in ["Status", "Count"]:
        table.add_column(header, justify="center")

    for s, count in status.items():
        table.add_row(status_map[s], str(count))

    console.print(table)


def show_machine_status() -> None:
    """
    Print the current machine status in Boulder Opal cloud environment for your organization.
    """

    status_map = {
        "Online": "online",
        "Starting up": "initializing",
        "Pending": "pending",
        "Shutting down": "terminating",
        "Offline": "offline",
    }

    status = get_cloud_router("Machine status").get_machine_status()

    table = Table(title="Machine status", show_lines=True)
    console = Console()

    for header in ["Status", "Count"]:
        table.add_column(header, justify="center")

    for state, key in status_map.items():
        number = status.get(key)
        if number is None:
            raise RuntimeError(f"Unknown machine state. Expected {list(status.keys())}, got {key}.")
        table.add_row(state, str(number))

    console.print(table)



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Module to host some utils functions, which are exposed at the top-level.
"""

from __future__ import annotations

from qctrlworkflowclient.utils import package_versions_table


def print_package_versions() -> None:
    """
    Print a Markdown-formatted table showing the Python version being used,
    as well as the versions of some loaded packages that are relevant to Boulder Opal.
    """

    package_names = [
        # External packages.
        "jsonpickle",
        "matplotlib",
        "mloop",
        "numpy",
        "qiskit",
        "qutip",
        "scipy",
        # Q-CTRL packages.
        "boulderopal",
        "qctrlcommons",
        "qctrlexperimentscheduler",
        "qctrlmloop",
        "qctrlopencontrols",
        "qctrlqua",
        "qctrlvisualizer",
        "qctrlworkflowclient",
    ]

    print(package_versions_table(package_names))


__all__ = ["print_package_versions"]



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from boulderopal._validation.basic import (
    ArrayT,
    ScalarT,
    nullable,
    validate_enum,
)
from boulderopal._validation.exceptions import Checker
from boulderopal._validation.pydantic import (
    pipe,
    sequence_like,
    type_pipe,
    validated,
)

__all__ = [
    "ArrayT",
    "Checker",
    "ScalarT",
    "nullable",
    "pipe",
    "sequence_like",
    "type_pipe",
    "validated",
    "validate_enum",
]



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from collections import namedtuple
from enum import Enum
from functools import partial
from operator import (
    ge,
    gt,
    le,
    lt,
    ne,
)
from typing import (
    Any,
    Callable,
    Optional,
    Type,
    TypeVar,
    Union,
)

import numpy as np

from boulderopal._typing import (
    Concatenate,
    ParamSpec,
    Self,
)
from boulderopal._validation.exceptions import Checker

T = TypeVar("T", bound=Enum)


def validate_enum(enum_: Type[T], item: T | str) -> str:
    """
    Check whether the item is a valid option in enum_. If so, return the name of that option.
    Otherwise, raise an error.

    Parameters
    ----------
    enum_ : T
        An Enum where we expect the option value to be a str.
    item : T or str
        The item to be checked with enum_.

    Returns
    -------
    str
        The name of a valid enum option.
    """
    if isinstance(item, enum_):
        return item.name
    try:
        return getattr(enum_, item).name  # type: ignore
    except (TypeError, AttributeError) as err:
        raise ValueError(
            f"Only the following options are allowed: {list(enum_.__members__)}, got {item}.",
        ) from err


def _filter_0d_array(val: Any) -> Any:
    if isinstance(val, np.ndarray) and val.ndim == 0:
        return val.item()
    return val


def _is_integer(val: Any) -> bool:
    return isinstance(_filter_0d_array(val), ScalarT.INT.value.types)


def _is_real(val: Any) -> bool:
    return _is_integer(val) or isinstance(_filter_0d_array(val), ScalarT.REAL.value.types)


def _is_complex(val: Any) -> bool:
    return _is_real(val) or _is_strict_complex(val)


def _is_strict_complex(val: Any) -> bool:
    return isinstance(_filter_0d_array(val), ScalarT.COMPLEX.value.types)


def _is_numeric(val: Any) -> bool:
    return _is_real(val) or _is_complex(val)


def _number_converter(val: Any) -> int | float | complex:
    if _is_integer(val):
        return int(val)
    if _is_real(val):
        return float(val)
    return complex(val)


# types: supported types defined by Python and Numpy for a given dtype
# checker: a callable to check the input scalar
# converter: a callable to convert the scalar to the corresponding Python primitive type
_ScalarTValidator = namedtuple("_ScalarTValidator", ["types", "checker", "converter"])

_Number = Union[int, float, complex, np.integer, np.float64, np.complex128]
_SCALAR = TypeVar("_SCALAR", bound=_Number)

_CompareOp = namedtuple("_CompareOp", ["op", "message"])


class _Compare(Enum):
    LESS = _CompareOp(lt, "smaller than")
    LESS_EQUAL = _CompareOp(le, "smaller than or equal to")
    GREATER = _CompareOp(gt, "greater than")
    GREATER_EQUAL = _CompareOp(ge, "greater than or equal to")
    NOT_EQUAL = _CompareOp(ne, "not equal to")


class _ScalarValidator:
    def __init__(self, type_: ScalarT, name: str):
        self._type = type_
        self._name = name
        self._validator_pipe: list[Callable[[_Number], None]] = []

    def _compare(self, operator: _Compare, rhs: float | int, name: str, value: _Number) -> None:
        Checker.VALUE(
            not _is_strict_complex(value),
            f"Complex number {self._name} doesn't support comparison.",
        )
        _name_val = str(rhs) if len(name) == 0 else f"{name} {rhs}"
        Checker.VALUE(
            operator.value.op(value, rhs),
            f"The {self._name} must be {operator.value.message} {_name_val}.",
            {self._name: value},
        )

    def lt(self, bound: float | int, rhs_name: str = "") -> Self:
        """
        Scalar < bound.
        """
        self._validator_pipe.append(partial(self._compare, _Compare.LESS, bound, rhs_name))
        return self

    def le(self, bound: float | int, rhs_name: str = "") -> Self:
        """
        Scalar <= bound.
        """
        self._validator_pipe.append(partial(self._compare, _Compare.LESS_EQUAL, bound, rhs_name))
        return self

    def gt(self, bound: float | int, rhs_name: str = "") -> Self:
        """
        Scalar > bound.
        """
        self._validator_pipe.append(partial(self._compare, _Compare.GREATER, bound, rhs_name))
        return self

    def ge(self, bound: float | int, rhs_name: str = "") -> Self:
        """
        Scalar >= bound.
        """
        self._validator_pipe.append(partial(self._compare, _Compare.GREATER_EQUAL, bound, rhs_name))
        return self

    def ne(self, value: float | int, rhs_name: str = "") -> Self:
        """
        Scalar != value.
        """
        self._validator_pipe.append(partial(self._compare, _Compare.NOT_EQUAL, value, rhs_name))
        return self

    def __call__(self, value: _SCALAR, *, name: Optional[str] = None) -> _SCALAR:
        if name is not None:
            self._name = name
        Checker.TYPE(
            self._type.value.checker(value),
            f"The {self._name} must be a {self._type.name.lower()}.",
            {self._name: value},
        )
        _converted = self._type.value.converter(value)
        for _func in self._validator_pipe:
            _func(_converted)
        return _converted


class ScalarT(Enum):
    """
    Store dtypes to validate both Python and NumPy types.
    """

    INT = _ScalarTValidator((int, np.integer), _is_integer, int)
    REAL = _ScalarTValidator((float, np.float64), _is_real, float)
    COMPLEX = _ScalarTValidator((complex, np.complex128), _is_complex, complex)
    NUMERIC = _ScalarTValidator(None, _is_numeric, _number_converter)

    def __call__(self, name: str = "") -> _ScalarValidator:
        """
        Return a `_ScalarValidator` to validate a scalar by the dtype with optional constraints.

        Note that you can ignore the `name` argument and set it later when invoking the validator.
        """
        return _ScalarValidator(self, name)


# valid_dtype_kinds: valid values for the array's dtype.kind.
# dtype: data type of the returned validated NumPy array
_ArrayTValidator = namedtuple("_ArrayTValidator", ["valid_dtype_kinds", "dtype"])


def _no_comparison(value: np.ndarray, name: str) -> None:
    Checker.VALUE(value.dtype.kind != "c", f"Complex array {name} doesn't support comparison.")


class _ArrayValidator:
    def __init__(self, type_: ArrayT, name: str):
        self._type = type_
        self._name = name
        self._validator_pipe: list[Callable[[np.ndarray], tuple[bool | np.bool_, str]]] = []

    def ndim(self, dim_: int) -> Self:
        """
        The number of dimensions of the array must be dim_.
        """
        self._validator_pipe.append(
            lambda x: (x.ndim == dim_, f"The {self._name} must be a {dim_}D array."),
        )
        return self

    def shape(self, shape_: tuple, extra: str = "") -> Self:
        """
        The shape of the array must be shape_.
        """

        _shape_message = f"{shape_}" if len(extra) == 0 else f"{extra}, {shape_}"
        message = f"The shape of {self._name} must be {_shape_message}."
        self._validator_pipe.append(lambda x: (x.shape == shape_, message))
        return self

    def no_scalar(self) -> Self:
        """
        Do not allow scalar array.
        """
        self._validator_pipe.append(
            lambda x: (
                len(x.shape) > 0,
                f"The {self._name} must not be a scalar array.",
            ),
        )
        return self

    def _compare(
        self,
        operator: _Compare,
        rhs: float | int,
        name: str,
        value: np.ndarray,
    ) -> tuple[np.bool_, str]:
        _no_comparison(value, name)
        _name_val = str(rhs) if len(name) == 0 else f"{name} {rhs}"
        return (
            np.all(operator.value.op(value, rhs)),
            f"The value in {self._name} must be {operator.value.message} {_name_val}.",
        )

    def lt(self, bound: float | int, rhs_name: str = "") -> Self:
        """
        np.all(array < bound).
        """
        self._validator_pipe.append(partial(self._compare, _Compare.LESS, bound, rhs_name))
        return self

    def le(self, bound: float | int, rhs_name: str = "") -> Self:
        """
        np.all(array <= bound).
        """
        self._validator_pipe.append(partial(self._compare, _Compare.LESS_EQUAL, bound, rhs_name))
        return self

    def gt(self, bound: float | int, rhs_name: str = "") -> Self:
        """
        np.all(array > bound).
        """
        self._validator_pipe.append(partial(self._compare, _Compare.GREATER, bound, rhs_name))
        return self

    def ge(self, bound: float | int, rhs_name: str = "") -> Self:
        """
        np.all(array >= bound).
        """
        self._validator_pipe.append(partial(self._compare, _Compare.GREATER_EQUAL, bound, rhs_name))
        return self

    def ascend(self) -> Self:
        """
        In ascending order.
        """

        def _check(x: np.ndarray) -> tuple[np.bool_, str]:
            _no_comparison(x, self._name)
            return (
                np.all(np.diff(x) > 0),
                f"The {self._name} must be in ascending order and not have duplicate values.",
            )

        self._validator_pipe.append(_check)
        return self

    def descend(self) -> Self:
        """
        In descending order.
        """

        def _check(x: np.ndarray) -> tuple[np.bool_, str]:
            _no_comparison(x, self._name)
            return (
                np.all(np.diff(x) < 0),
                f"The {self._name} must be in descending order and not have duplicate values.",
            )

        self._validator_pipe.append(_check)
        return self

    def square(self) -> Self:
        """
        Last two dimensions are equal.
        """
        self._validator_pipe.append(
            lambda x: (
                x.shape[-1] == x.shape[-2],
                f"The {self._name} must be square in the last two dimensions.",
            ),
        )
        return self

    def hermitian(self) -> Self:
        """
        Equals its adjoint.
        """
        self._validator_pipe.append(
            lambda x: (
                np.allclose(x, np.moveaxis(x, -1, -2).conj()),
                f"The {self._name} must be Hermitian.",
            ),
        )
        return self

    def idempotent(self) -> Self:
        """
        Equals its square.
        """
        self._validator_pipe.append(
            lambda x: (
                np.allclose(x, x @ x),
                f"The {self._name} must be idempotent, but does not equal its square.",
            ),
        )
        return self

    def __call__(self, value: Any, *, name: Optional[str] = None) -> np.ndarray:
        if name is not None:
            self._name = name
        try:
            _converted = np.asarray(value)
        except ValueError as exc:
            raise TypeError(
                f"Can't convert {self._name} to an array, {self._name}={value}.",
            ) from exc
        Checker.VALUE(_converted.size > 0, f"The {self._name} must not be an empty array.")
        Checker.TYPE(
            _converted.dtype.kind in self._type.value.valid_dtype_kinds,
            f"The {self._name} must be a {self._type.name.lower()} array.",
        )

        # Special handler for cases:
        # - int typed array to address the signed and unsigned cases.
        # - a generic numeric array
        # That is, we preserve the dtype from users in these cases.
        if self._type not in (ArrayT.INT, ArrayT.NUMERIC):
            try:
                _converted = _converted.astype(self._type.value.dtype, casting="safe")
            except TypeError as err:
                raise TypeError(
                    f"Expected {self._name} as an array of {self._type.name.lower()} dtype, "
                    f"but got {_converted.dtype}.",
                ) from err

        for func in self._validator_pipe:
            Checker.VALUE(*func(_converted))

        return _converted


class ArrayT(Enum):
    """
    Store dtypes to validate array-likes.
    """

    INT = _ArrayTValidator("iu", np.integer)
    REAL = _ArrayTValidator("iuf", np.float64)
    COMPLEX = _ArrayTValidator("iufc", np.complex128)
    NUMERIC = _ArrayTValidator("iufc", None)

    def __call__(self, name: str = "") -> _ArrayValidator:
        """
        Return an `_ArrayValidator` to validate an array-like by the dtype
        with optional constraints.

        Note that you can ignore the `name` argument and set it later when invoking the validator.
        """
        return _ArrayValidator(self, name)


_VAL = TypeVar("_VAL")
P = ParamSpec("P")


def nullable(
    validator: Callable[Concatenate[_VAL, P], _VAL],
    value: Optional[_VAL],
    *args: P.args,
    **kwargs: P.kwargs,
) -> Optional[_VAL]:
    """
    Validate a parameter that can be None.

    When the parameter holds a non-null value, the validator callable is used to check the value.
    The validator takes the value as the first argument and some other options as defined by P, it
    returns the same type as the input value (strictly speaking, the returned type is something
    that can be converted from the input one. But in reality, we expect them to be
    interchangeable). The P annotation here allows mypy to also check the types for the
    resting arguments of the validator.
    """
    if value is None:
        return value
    return validator(value, *args, **kwargs)
